# -*- coding: utf-8 -*-

# Read more about operations with processes by win32 api here:
# https://learn.microsoft.com/en-us/windows/win32/api/memoryapi/
# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/
# https://learn.microsoft.com/en-us/windows/win32/api/psapi/
# ...

from ..enums import ScanTypesEnum
from ..util import convert_from_byte_array, get_c_type_of, scan_memory, scan_memory_for_exact_value

from .enums import MemoryAllocationStatesEnum, MemoryProtectionsEnum, MemoryTypesEnum
from .types import MEMORY_BASIC_INFORMATION, SYSTEM_INFO, WNDENUMPROC

from typing import Dict, Generator, Optional, Sequence, Tuple, Type, TypeVar, Union

import ctypes
import ctypes.wintypes

# Load the libraries.
kernel32 = ctypes.windll.LoadLibrary("kernel32.dll")
user32 = ctypes.windll.LoadLibrary("user32.dll")

# Set the argtypes to prevent ArgumentError.
kernel32.VirtualQueryEx.argtypes = (
    ctypes.wintypes.HANDLE, ctypes.wintypes.LPCVOID, ctypes.POINTER(MEMORY_BASIC_INFORMATION), ctypes.c_uint32
)


# Get the user's system information.
system_information = SYSTEM_INFO()
kernel32.GetSystemInfo(ctypes.byref(system_information))


T = TypeVar("T")


def CloseProcessHandle(process_handle: int) -> int:
    """
    Close the process handle.
    """
    return kernel32.CloseHandle(process_handle)


def GetMemoryRegions(process_handle: int) -> Generator[dict, None, None]:
    """
    Generates dictionaries with the address and size of a region used by the process.
    """
    mem_region_begin = system_information.lpMinimumApplicationAddress
    mem_region_end = system_information.lpMaximumApplicationAddress

    current_address = mem_region_begin

    while current_address < mem_region_end:
        region = MEMORY_BASIC_INFORMATION()
        kernel32.VirtualQueryEx(process_handle, current_address, ctypes.byref(region), ctypes.sizeof(region))

        yield {"address": current_address, "size": region.RegionSize, "struct": region}

        current_address += region.RegionSize


def GetProcessHandle(access_right: int, inherit: bool, pid: int) -> int:
    """
    Get a process ID and return its process handle.

    :param access_right: The access to the process object. This access right is
    checked against the security descriptor for the process. This parameter can
    be one or more of the process access rights.

    :param inherit: if this value is TRUE, processes created by this process
    will inherit the handle. Otherwise, the processes do not inherit this handle.

    :param pid: The identifier of the local process to be opened.
    """
    return kernel32.OpenProcess(access_right, inherit, pid)


def GetProcessIdByWindowTitle(window_title: str) -> int:
    """
    Return the process ID by querying a window title.
    """
    result = ctypes.c_uint32(0)

    string_buffer_size = len(window_title) + 2  # (+2) for the next possible character of a title and the NULL char.
    string_buffer = ctypes.create_unicode_buffer(string_buffer_size)

    def callback(hwnd, size):
        """
        This callback is used to get a window handle and compare
        its title with the target window title.

        To continue enumeration, the callback function must return TRUE;
        to stop enumeration, it must return FALSE.
        """
        nonlocal result, string_buffer

        user32.GetWindowTextW(hwnd, string_buffer, size)

        # Compare the window titles and get the process ID.
        if window_title == string_buffer.value:
            user32.GetWindowThreadProcessId(hwnd, ctypes.byref(result))
            return False

        # Indicate it must continue enumeration.
        return True

    # Enumerates all top-level windows on the screen by passing the handle to each window,
    # in turn, to an application-defined callback function.
    user32.EnumWindows(WNDENUMPROC(callback), string_buffer_size)

    return result.value


def ReadProcessMemory(
    process_handle: int,
    address: int,
    pytype: Type[T],
    bufflength: int
) -> T:
    """
    Return a value from a memory address.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    data = get_c_type_of(pytype, bufflength)
    kernel32.ReadProcessMemory(process_handle, ctypes.c_void_p(address), ctypes.byref(data), bufflength, None)

    return data.value.decode() if pytype is str else data.value


def SearchAddressesByValue(
    process_handle: int,
    pytype: Type[T],
    bufflength: int,
    value: Union[bool, int, float, str, bytes, tuple],
    scan_type: ScanTypesEnum = ScanTypesEnum.EXACT_VALUE,
    progress_information: bool = False,
    writeable_only: bool = False,
) -> Generator[Union[int, Tuple[int, dict]], None, None]:
    """
    Search the whole memory space, accessible to the process,
    for the provided value, returning the found addresses.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    # Convert the target value, or all values of a tuple, as bytes.
    target_values = value if isinstance(value, tuple) else (value,)

    conversion_buffer = list()

    for v in target_values:
        target_value = get_c_type_of(pytype, bufflength)
        target_value.value = v.encode() if isinstance(v, str) else v

        target_value_bytes = ctypes.cast(ctypes.byref(target_value), ctypes.POINTER(ctypes.c_byte * bufflength))
        conversion_buffer.append(bytes(target_value_bytes.contents))

    target_value_bytes = tuple(conversion_buffer) if isinstance(value, tuple) else conversion_buffer[0]

    # Get the memory regions, computing the total amount of memory to be scanned.
    checked_memory_size = 0
    memory_total = 0
    memory_regions = list()

    for region in GetMemoryRegions(process_handle):

        # Only committed, non-shared and readable memory pages.
        if region["struct"].State != MemoryAllocationStatesEnum.MEM_COMMIT.value: continue
        if region["struct"].Type != MemoryTypesEnum.MEM_PRIVATE.value: continue
        if region["struct"].Protect & MemoryProtectionsEnum.PAGE_READABLE.value == 0: continue

        # If writeable_only is True, checks if the memory page is writeable.
        if writeable_only and region["struct"].Protect & MemoryProtectionsEnum.PAGE_READWRITEABLE.value == 0: continue

        memory_total += region["size"]
        memory_regions.append(region)

    # Sort the list to return ordered addresses.
    memory_regions.sort(key=lambda region: region["address"])

    # Check each memory region used by the process.
    for region in memory_regions:
        address, size = region["address"], region["size"]
        region_data = (ctypes.c_byte * size)()

        # Get data from the region.
        kernel32.ReadProcessMemory(process_handle, ctypes.c_void_p(address), ctypes.byref(region_data), size, None)

        # Choose the searching method.
        searching_method = scan_memory

        if scan_type in [ScanTypesEnum.EXACT_VALUE, ScanTypesEnum.NOT_EXACT_VALUE]:
            searching_method = scan_memory_for_exact_value

        # Search the value and return the found addresses.
        for offset in searching_method(region_data, size, target_value_bytes, bufflength, scan_type, pytype is str):
            found_address = address + offset

            extra_information = {
                "memory_total": memory_total,
                "progress": (checked_memory_size + offset) / memory_total,
            }
            yield (found_address, extra_information) if progress_information else found_address

        # Compute the region size to the checked memory size.
        checked_memory_size += size


def SearchValuesByAddresses(
    process_handle: int,
    pytype: Type[T],
    bufflength: int,
    addresses: Sequence[int],
    *,
    memory_regions: Optional[Sequence[Dict]] = None,
    raise_error: bool = False,
) -> Generator[Tuple[int, Optional[T]], None, None]:
    """
    Search the whole memory space, accessible to the process,
    for the provided list of addresses, returning their values.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    memory_regions = list(memory_regions) if memory_regions else list()
    addresses = sorted(addresses)

    # If no memory page has been given, get all committed, non-shared and readable memory pages.
    if not memory_regions:
        for region in GetMemoryRegions(process_handle):
            if region["struct"].State != MemoryAllocationStatesEnum.MEM_COMMIT.value: continue
            if region["struct"].Type != MemoryTypesEnum.MEM_PRIVATE.value: continue
            if region["struct"].Protect & MemoryProtectionsEnum.PAGE_READABLE.value == 0: continue

            memory_regions.append(region)

    memory_regions.sort(key=lambda region: region["address"])
    address_index = 0

    # Walk by each memory region.
    for region in memory_regions:
        if address_index >= len(addresses): break

        target_address = addresses[address_index]

        # Check if the memory region contains the target address.
        base_address, size = region["address"], region["size"]
        if not (base_address <= target_address < base_address + size): continue

        region_data = (ctypes.c_byte * size)()

        # Get data from the region.
        kernel32.ReadProcessMemory(process_handle, ctypes.c_void_p(base_address), ctypes.byref(region_data), size, None)

        # Get the value of each address.
        while base_address <= target_address < base_address + size:
            offset = target_address - base_address
            address_index += 1

            try:
                data = region_data[offset: offset + bufflength]
                data = (ctypes.c_byte * bufflength)(*data)
                yield target_address, convert_from_byte_array(data, pytype, bufflength)

            except Exception as error:
                if raise_error: raise error
                yield target_address, None

            if address_index >= len(addresses): break
            target_address = addresses[address_index]


def WriteProcessMemory(
    process_handle: int,
    address: int,
    pytype: Type[T],
    bufflength: int,
    value: Union[bool, int, float, str, bytes]
) -> T:
    """
    Write a value to a memory address.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    data = get_c_type_of(pytype, bufflength)
    data.value = value.encode() if isinstance(value, str) else value

    kernel32.WriteProcessMemory(process_handle, ctypes.c_void_p(address), ctypes.byref(data), bufflength, None)

    return value



# -*- coding: utf-8 -*-

# Read more about proc and memory mapping here:
# https://man7.org/linux/man-pages/man5/proc.5.html

# Read more about iovec here:
# https://man7.org/linux/man-pages/man3/iovec.3type.html

from ctypes import Structure, c_char_p, c_size_t, c_uint, c_void_p


class MEMORY_BASIC_INFORMATION(Structure):
    _fields_ = [
        ("BaseAddress", c_uint),
        ("RegionSize", c_uint),
        ("Privileges", c_char_p),
        ("Offset", c_uint),
        ("MajorID", c_uint),
        ("MinorID", c_uint),
        ("InodeID", c_uint),
        ("Path", c_char_p),
    ]


class iovec(Structure):
    """
    Describes a region of memory, beginning at iov_base address and
    with the size of iov_len bytes.  System calls use arrays of this
    structure, where each element of the array represents a memory
    region, and the whole array represents a vector of memory
    regions. The maximum number of iovec structures in that array is
    limited by IOV_MAX (defined in <limits.h>, or accessible via the
    call sysconf(_SC_IOV_MAX)).

    Reference: https://man7.org/linux/man-pages/man3/iovec.3type.html
    """
    _fields_ = [
        ("iov_base", c_void_p),
        ("iov_len", c_size_t)
    ]



# -*- coding: utf-8 -*-

from enum import Enum


class PtraceCommandsEnum(Enum):
    """
    Enum with commands for ptrace() system call.

    Read more about ptrace commands here:
    https://man7.org/linux/man-pages/man2/ptrace.2.html
    """
    # Turns the calling thread into a tracee. The thread continues to
    # run (doesn't enter ptrace-stop). A common practice is to follow
    # the PTRACE_TRACEME with "raise(SIGSTOP);" and allow the parent,
    # which is our tracer now, to observe our signal-delivery-stop.
    PTRACE_TRACEME = 0

    # PEEKTEXT and PEEKDATE read a word at the address addr in the
    # tracee's memory, returning the word as the result of the ptrace()
    # call. Linux does not have separate text and data address spaces,
    # so these two requests are currently equivalent.
    PTRACE_PEEKTEXT = 1
    PTRACE_PEEKDATA = 2

    # Read a word at offset addr in the tracee's USER area, which holds
    # the registers and other information about the process. The word is
    # returned as the result of the ptrace() call. Typically, the offset
    # must be word-aligned, though this might vary by architecture.
    PTRACE_PEEKUSER = 3

    # POKETEXT and POKEDATA copy the word data to the address addr in the
    # tracee's memory. These two requests are currently equivalent.
    PTRACE_POKETEXT = 4
    PTRACE_POKEDATA = 5

    # Copy the word data to offset addr in the tracee's USER area. As
    # for PTRACE_PEEKUSER, the offset must typically be word-aligned. In
    # order to maintain the integrity of the kernel, some modifications
    # to the USER area are disallowed.
    PTRACE_POKEUSER = 6

    # Restart the stopped tracee process. If data is nonzero, it is
    # interpreted as the number of a signal to be delivered to the tracee;
    # otherwise, no signal is delivered. Thus, for example, the tracer can
    # control whether a signal sent to the tracee is delivered or not.
    PTRACE_CONT = 7

    # Send the tracee a SIGKILL to terminate it. This operation is deprecated;
    # do not use it! Instead, send a SIGKILL directly using kill(2) or tgkill(2).
    # The problem with PTRACE_KILL is that it requires the tracee to be in
    # signal-delivery-stop, otherwise it may not work (i.e., may complete
    # successfully but won't kill the tracee). By contrast, sending a SIGKILL
    # directly has no such limitation.
    PTRACE_KILL = 8

    # GETREGS and GETFPREGS copy the tracee's general-purpose or floating-point
    # registers, respectively, to the address data in the tracer. Note that SPARC
    # systems have the meaning of data and addr reversed; that is, data is ignored
    # and the registers are copied to the address addr. PTRACE_GETREGS and
    # PTRACE_GETFPREGS are not present on all architectures.
    PTRACE_GETREGS = 12
    PTRACE_GETFPREGS = 14

    # SETREGS and SETFPREGS modify the tracee's general-purpose or floating-point
    # registers, respectively, from the address data in the tracer. As for
    # PTRACE_POKEUSER, some general-purpose register modifications may be
    # disallowed. Note that SPARC systems have the meaning of data and addr
    # reversed; that is, data is ignored and the registers are copied from the
    # address addr. PTRACE_SETREGS and PTRACE_SETFPREGS are not present on all
    # architectures.
    PTRACE_SETREGS = 13
    PTRACE_SETFPREGS = 15

    # Attach to the process specified in pid, making it a tracee of the calling
    # process. The tracee is sent a SIGSTOP, but will not necessarily have
    # stopped by the completion of this call; use waitpid(2) to wait for the
    # tracee to stop. See the "Attaching and detaching" subsection for additional
    # information. Permission to perform a PTRACE_ATTACH is governed by a ptrace
    # access mode PTRACE_MODE_ATTACH_REALCREDS check.
    PTRACE_ATTACH = 16

    # Restart the stopped tracee as for PTRACE_CONT, but first detach from it.
    # Under Linux, a tracee can be detached in this way regardless of which
    # method was used to initiate tracing.
    PTRACE_DETACH = 17

    # SINGLESTEP and SYSCALL restart the stopped tracee as for PTRACE_CONT,
    # but arrange for the tracee to be stopped at the next entry to or exit
    # from a system call, or after execution of a single instruction,
    # respectively. The tracee will also, as usual, be stopped upon receipt
    # of a signal. From the tracer's perspective, the tracee will appear to
    # have been stopped by receipt of a SIGTRAP. So, for PTRACE_SYSCALL, for
    # example, the idea is to inspect the arguments to the system call at the
    # first stop, then do another PTRACE_SYSCALL and inspect the return value
    # of the system call at the second stop. The data argument is treated as
    # for PTRACE_CONT.
    PTRACE_SINGLESTEP = 9
    PTRACE_SYSCALL = 24

    # Set ptrace options from data. Data is interpreted as a bit mask of options,
    # which are specified by the following flags:
    # - PTRACE_O_EXITKILL
    # - PTRACE_O_TRACECLONE
    # - PTRACE_O_TRACEFORK
    # - PTRACE_O_TRACESYSGOOD
    # - PTRACE_O_TRACEVFORK
    # - PTRACE_O_TRACEVFORKDONE
    # - PTRACE_O_TRACESECCOMP
    # - PTRACE_O_SUSPEND_SECCOMP
    PTRACE_SETOPTIONS = 0x4200



# -*- coding: utf-8 -*-

from ..enums import ScanTypesEnum
from .search.kmp import KMPSearch

from typing import Generator, Sequence, Tuple, Union
import ctypes
import sys


def scan_memory_for_exact_value(
    memory_region_data: Sequence,
    memory_region_data_size: int,
    target_value: bytes,
    target_value_size: int,
    comparison: ScanTypesEnum = ScanTypesEnum.EXACT_VALUE,
    *args, **kwargs
) -> Generator[int, None, None]:
    """
    Search for an exact value at the memory region.

    This method uses an efficient searching algorithm.
    """
    searcher = KMPSearch(target_value, target_value_size)
    last_index = 0

    for found_index in searcher.search(memory_region_data, memory_region_data_size):

        # Return the found index if user is searching for an exact value.
        if comparison is ScanTypesEnum.EXACT_VALUE:
            yield found_index
            continue

        # Return the interval between last_index and found_address, if user is searching for a different value.
        for different_index in range(last_index, found_index):
            yield different_index
        last_index = found_index + 1

    # If user is searching for a different value, return the rest of the addresses that were not found.
    if comparison is ScanTypesEnum.NOT_EXACT_VALUE:
        for different_index in range(last_index, memory_region_data_size):
            yield different_index


def scan_memory(
    memory_region_data: Sequence,
    memory_region_data_size: int,
    target_value: Union[bytes, Tuple[bytes]],
    target_value_size: int,
    scan_type: ScanTypesEnum,
    is_string: bool,
) -> Generator[int, None, None]:
    """
    Search for a value at the memory region.
    """
    byte_order = sys.byteorder if not is_string else "big"

    # If target_value is a tuple, it means the user wants to compare to more than one value.
    if isinstance(target_value, tuple):
        start_target_value_int = int.from_bytes(target_value[0], byte_order)
        end_target_value_int = int.from_bytes(target_value[1], byte_order)
    else: target_value_int = int.from_bytes(target_value, byte_order)

    for found_index in range(memory_region_data_size - target_value_size):

        # Convert data to an integer.
        data = memory_region_data[found_index: found_index + target_value_size]
        data = bytes((ctypes.c_byte * target_value_size)(*data))
        data = int.from_bytes(data, byte_order)

        # Compare value between.
        if scan_type is ScanTypesEnum.VALUE_BETWEEN and (start_target_value_int > data or data > end_target_value_int): continue
        elif scan_type is ScanTypesEnum.NOT_VALUE_BETWEEN and (start_target_value_int < data < end_target_value_int): continue

        # Compare the value.
        elif scan_type is ScanTypesEnum.EXACT_VALUE and data != target_value_int: continue
        elif scan_type is ScanTypesEnum.NOT_EXACT_VALUE and data == target_value_int: continue
        elif scan_type is ScanTypesEnum.BIGGER_THAN and data <= target_value_int: continue
        elif scan_type is ScanTypesEnum.SMALLER_THAN and data >= target_value_int: continue
        elif scan_type is ScanTypesEnum.BIGGER_THAN_OR_EXACT_VALUE and data < target_value_int: continue
        elif scan_type is ScanTypesEnum.SMALLER_THAN_OR_EXACT_VALUE and data > target_value_int: continue

        yield found_index



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Generator, Optional, Sequence, Tuple, Type, TypeVar, Union

from ..enums import ScanTypesEnum
from ..process.info import ProcessInfo


T = TypeVar("T")


class AbstractProcess(ABC):
    """
    Abstract class to represent a process.
    """

    @abstractmethod
    def __init__(self, *, window_title: Optional[str] = None, process_name: Optional[str] = None, pid: Optional[int] = None):
        """
        :param window_title: window title of the target program.
        :param process_name: name of the target process.
        :param pid: process ID.
        """
        self._process_info = ProcessInfo()

        # Set the attributes to the process.
        if pid:
            self._process_info.pid = pid

        elif window_title:
            self._process_info.window_title = window_title

        elif process_name:
            self._process_info.process_name = process_name

        else:
            raise TypeError("You must pass an argument to one of these parameters (window_title, process_name, pid).")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    @property
    def pid(self) -> int:
        return self._process_info.pid

    @abstractmethod
    def close(self) -> bool:
        """
        Close the process handle.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_memory_regions(self) -> Generator[dict, None, None]:
        """
        Generates dictionaries with the address, size and other
        information of each memory region used by the process.
        """
        raise NotImplementedError()

    @abstractmethod
    def search_by_addresses(
        self,
        pytype: Type[T],
        bufflength: int,
        addresses: Sequence[int],
        *,
        raise_error: bool = False,
    ) -> Generator[Tuple[int, Optional[T]], None, None]:
        """
        Search the whole memory space, accessible to the process,
        for the provided list of addresses, returning their values.
        """
        raise NotImplementedError()

    @abstractmethod
    def search_by_value(
        self,
        pytype: Type[T],
        bufflength: int,
        value: Union[bool, int, float, str, bytes],
        scan_type: ScanTypesEnum = ScanTypesEnum.EXACT_VALUE,
        *,
        progress_information: bool = False,
        writeable_only: bool = False,
    ) -> Generator[Union[int, Tuple[int, dict]], None, None]:
        """
        Search the whole memory space, accessible to the process,
        for the provided value, returning the found addresses.

        :param pytype: type of value to be queried (bool, int, float, str or bytes).
        :param bufflength: value size in bytes (1, 2, 4, 8).
        :param value: value to be queried (bool, int, float, str or bytes).
        :param scan_type: the way to compare the values.
        :param progress_information: if True, a dictionary with the progress information will be return.
        :param writeable_only: if True, search only at writeable memory regions.
        """
        raise NotImplementedError()

    def search_by_value_between(
        self,
        pytype: Type[T],
        bufflength: int,
        start: Union[bool, int, float, str, bytes],
        end: Union[bool, int, float, str, bytes],
        *,
        not_between: bool = False,
        progress_information: bool = False,
        writeable_only: bool = False,
    ) -> Generator[Union[int, Tuple[int, dict]], None, None]:
        """
        Search the whole memory space, accessible to the process,
        for a value within the provided range, returning the found addresses.

        :param pytype: type of value to be queried (bool, int, float, str or bytes).
        :param bufflength: value size in bytes (1, 2, 4, 8).
        :param start: minimum inclusive value to be queried (bool, int, float, str or bytes).
        :param end: maximum inclusive value to be queried (bool, int, float, str or bytes).
        :param not_between: if True, return only addresses of values that are NOT within the range.
        :param progress_information: if True, a dictionary with the progress information will be return.
        :param writeable_only: if True, search only at writeable memory regions.
        """
        raise NotImplementedError()

    @abstractmethod
    def read_process_memory(
        self,
        address: int,
        pytype: Type[T],
        bufflength: int
    ) -> T:
        """
        Return a value from a memory address.

        :param address: target memory address (ex: 0x006A9EC0).
        :param pytype: type of the value to be received (bool, int, float, str or bytes).
        :param bufflength: value size in bytes (1, 2, 4, 8).
        """
        raise NotImplementedError()

    @abstractmethod
    def write_process_memory(
        self,
        address: int,
        pytype: Type[T],
        bufflength: int,
        value: Union[bool, int, float, str, bytes]
    ) -> T:
        """
        Write a value to a memory address.

        :param address: target memory address (ex: 0x006A9EC0).
        :param pytype: type of value to be written into memory (bool, int, float, str or bytes).
        :param bufflength: value size in bytes (1, 2, 4, 8).
        :param value: value to be written (bool, int, float, str or bytes).
        """
        raise NotImplementedError()



from package import OpenProcess, ScanTypesEnum, version
from os import getpid
from typing import Optional
import ctypes
import platform
import random

print("Testing PyMemoryEditor version %s." % version)

print("\nOS Information: {} - {} {}".format(platform.platform(), *platform.architecture()[::-1]))
print("Processor Information: {} | {}\n".format(platform.machine(), platform.processor()))

process_id = getpid()
process: Optional[OpenProcess] = None


def generate_text(size):
    # Return a random text.
    return "".join([chr(random.randint(ord("A"), ord("Z"))) for letter in range(size)])


def test_open_process():
    global process

    # Open the process to write and read the process memory.
    process = OpenProcess(pid=process_id)


def test_read_bool():
    # Compare with True and False values.
    target_value_1 = ctypes.c_bool(True)
    target_value_2 = ctypes.c_bool(False)

    address_1 = ctypes.addressof(target_value_1)
    address_2 = ctypes.addressof(target_value_2)

    data_length = ctypes.sizeof(target_value_1)

    # Read the process memory and compare the results.
    result_1 = process.read_process_memory(address_1, bool, data_length)
    result_2 = process.read_process_memory(address_2, bool, data_length)

    assert type(result_1) is bool and result_1 == target_value_1.value
    assert type(result_2) is bool and result_2 == target_value_2.value


def test_read_float():
    # Get a random value to compare the result.
    target_value = ctypes.c_double(random.random())
    address = ctypes.addressof(target_value)
    data_length = ctypes.sizeof(target_value)

    # Read the process memory and compare the result.
    result = process.read_process_memory(address, float, data_length)
    assert type(result) is float and result == target_value.value


def test_read_int():
    # Get a random value to compare the result.
    target_value = ctypes.c_int(random.randint(0, 10000))
    address = ctypes.addressof(target_value)
    data_length = ctypes.sizeof(target_value)

    # Read the process memory and compare the result.
    result = process.read_process_memory(address, int, data_length)
    assert type(result) is int and result == target_value.value


def test_read_string():
    # Get a random text to compare the result.
    target_value = ctypes.create_string_buffer(20)
    target_value.value = generate_text(20).encode()

    address = ctypes.addressof(target_value)
    data_length = ctypes.sizeof(target_value)

    # Read the process memory and compare the result.
    result = process.read_process_memory(address, str, data_length)
    assert type(result) is str and result == target_value.value.decode()


def test_write_bool():
    # Compare with True and False values.
    original_value_1 = True
    original_value_2 = False

    new_value_1 = not original_value_1
    new_value_2 = not original_value_2

    target_value_1 = ctypes.c_bool(original_value_1)
    target_value_2 = ctypes.c_bool(original_value_2)

    address_1 = ctypes.addressof(target_value_1)
    address_2 = ctypes.addressof(target_value_2)

    data_length = ctypes.sizeof(target_value_1)

    # Write to the process memory and compare the results.
    process.write_process_memory(address_1, bool, data_length, new_value_1)
    process.write_process_memory(address_2, bool, data_length, new_value_2)

    assert target_value_1.value != original_value_1 and target_value_1.value == new_value_1
    assert target_value_2.value != original_value_2 and target_value_2.value == new_value_2


def test_write_float():
    # Get a random value to compare the result.
    original_value = random.random()
    new_value = original_value + 7651

    target_value = ctypes.c_double(original_value)
    address = ctypes.addressof(target_value)
    data_length = ctypes.sizeof(target_value)

    # Write to the process memory and compare the result.
    process.write_process_memory(address, float, data_length, new_value)
    assert target_value.value != original_value and target_value.value == new_value


def test_write_int():
    # Get a random value to compare the result.
    original_value = random.randint(0, 10000)
    new_value = original_value + 7651

    target_value = ctypes.c_int(original_value)
    address = ctypes.addressof(target_value)
    data_length = ctypes.sizeof(target_value)

    # Write to the process memory and compare the result.
    process.write_process_memory(address, int, data_length, new_value)
    assert target_value.value != original_value and target_value.value == new_value


def test_write_string():
    # Get a random text to compare the result.
    original_value = generate_text(20).encode()
    new_value = generate_text(20).encode()

    target_value = ctypes.create_string_buffer(20)
    target_value.value = original_value

    address = ctypes.addressof(target_value)
    data_length = ctypes.sizeof(target_value)

    # Write to the process memory and compare the result.
    process.write_process_memory(address, str, data_length, new_value.decode())
    assert target_value.value != original_value and target_value.value == new_value


def test_search_by_int_addresses():
    # Get random values to compare the result.
    test_length = 10

    target_values = [ctypes.c_int(random.randint(0, 10000)) for i in range(test_length)]
    data_length = ctypes.sizeof(target_values[0])

    target_values = {ctypes.addressof(v): v for v in target_values}
    addresses = list(target_values.keys())

    for address, value in process.search_by_addresses(int, data_length, addresses):
        assert target_values[address].value == value and type(value) is int


def test_search_by_float_addresses():
    # Get random values to compare the result.
    test_length = 10

    target_values = [ctypes.c_double(random.randint(0, 10000) / random.randint(0, 10000)) for i in range(test_length)]
    data_length = ctypes.sizeof(target_values[0])

    target_values = {ctypes.addressof(v): v for v in target_values}
    addresses = list(target_values.keys())

    for address, value in process.search_by_addresses(float, data_length, addresses):
        assert target_values[address].value == value and type(value) is float


def test_search_by_string_addresses():
    # Get random values to compare the result.
    string_length, test_length = 20, 10

    target_values = list()

    for i in range(test_length):
        value = ctypes.create_string_buffer(string_length)
        value.value = generate_text(string_length).encode()
        target_values.append(value)

    data_length = ctypes.sizeof(target_values[0])

    target_values = {ctypes.addressof(v): v for v in target_values}
    addresses = list(target_values.keys())

    for address, value in process.search_by_addresses(str, data_length, addresses):
        assert target_values[address].value.decode() == value and type(value) is str


def test_search_by_int():
    # Get random values to compare the result.
    test_length = 10

    target_values = [ctypes.c_int(random.randint(0, 10000)) for i in range(test_length)]
    addresses = [ctypes.addressof(v) for v in target_values]
    data_length = ctypes.sizeof(target_values[0])

    min_value = min([v.value for v in target_values])
    max_value = max([v.value for v in target_values])

    total = 0
    found = 0
    correct = 0

    # Get addresses of values exact or smaller than max_value.
    for found_address in process.search_by_value_between(int, data_length, min_value, max_value):

        # Check if the found address is a target address.
        if found_address in addresses:
            addresses.remove(found_address)
            found += 1

        total += 1

        # Check if the address really points to a valid value.
        value = process.read_process_memory(found_address, int, data_length)
        if min_value <= value <= max_value: correct += 1

    assert found / test_length >= 0.7
    assert correct / total >= 0.7  # Some of the addresses are beyond our control and may have their values changed.


def test_search_by_float():
    # Get random values to compare the result.
    test_length = 10

    target_values = [ctypes.c_double(random.randint(0, 10000)) for i in range(test_length)]
    addresses = [ctypes.addressof(v) for v in target_values]
    data_length = ctypes.sizeof(target_values[0])

    min_value = min([v.value for v in target_values])
    max_value = max([v.value for v in target_values])

    total = 0
    found = 0
    correct = 0

    # Get addresses of values exact or smaller than max_value.
    for found_address in process.search_by_value_between(float, data_length, min_value, max_value):

        # Check if the found address is a target address.
        if found_address in addresses:
            addresses.remove(found_address)
            found += 1

        total += 1

        # Check if the address really points to a valid value.
        value = process.read_process_memory(found_address, float, data_length)
        if min_value <= value <= max_value: correct += 1

    assert found / test_length >= 0.7
    assert correct / total >= 0.7  # Some of the addresses are beyond our control and may have their values changed.


def test_search_by_string():
    # Get random values to compare the result.
    string_length, test_length = 20, 10

    target_values = list()

    for i in range(test_length):
        value = ctypes.create_string_buffer(string_length)
        value.value = generate_text(string_length).encode()
        target_values.append(value)

    data_length = ctypes.sizeof(target_values[0])

    total = 0
    found = 0
    correct = 0

    # Get addresses of values exact or smaller than max_value.
    for target_value in target_values:
        for found_address in process.search_by_value(str, data_length, target_value.value, ScanTypesEnum.EXACT_VALUE):

            # Check if the found address is the target address.
            if found_address == ctypes.addressof(target_value):
                found += 1

            total += 1

            # Check if the address really points to a valid value.
            try:
                value = process.read_process_memory(found_address, str, data_length)
                if value == target_value.value.decode(): correct += 1
            except: pass

    assert found / test_length >= 0.7
    assert correct / total >= 0.7  # Some of the addresses are beyond our control and may have their values changed.


def test_search_by_string_between():
    # Get random values to compare the result.
    string_length, test_length = 20, 10

    values = list()

    for i in range(test_length * 2):
        value = ctypes.create_string_buffer(string_length)
        value.value = generate_text(string_length).encode()
        values.append(value)

    values.sort(key=lambda target_value: target_value.value)

    # Half of the set of strings is the target and the other half contains string that should be ignored by the scanner.
    target_values = [target_value for target_value in values[test_length // 4: test_length - test_length // 4]]

    addresses = [ctypes.addressof(v) for v in values]
    target_addresses = [ctypes.addressof(v) for v in target_values]

    data_length = ctypes.sizeof(target_values[0])

    min_value = target_values[0].value
    max_value = target_values[-1].value

    found = 0

    # Get addresses of values exact or smaller than max_value.
    for found_address in process.search_by_value_between(str, data_length, min_value, max_value):

        # Check if the found address is a target address.
        if found_address in target_addresses:
            addresses.remove(found_address)
            found += 1

        elif found_address in addresses:
            raise ValueError("Scanner returned the address of a clearly invalid string.")

    assert found / test_length >= 0.5


def test_close_process():
    # Try to close the process handle.
    assert process.close()



# -*- coding: utf-8 -*-
from enum import Enum


class MemoryAllocationStatesEnum(Enum):
    """
    Enum with all states of a memory page allocation.
    """
    # Indicates committed pages for which physical storage has been allocated,
    # either in memory or in the paging file on disk.
    MEM_COMMIT = 0x1000

    # Indicates free pages not accessible to the calling process and available
    # to be allocated. For free pages, the information in the AllocationBase,
    # AllocationProtect, Protect, and Type members is undefined.
    MEM_FREE = 0x10000

    # Allocates memory using large page support. The size and alignment must be a multiple
    # of the large-page minimum. To obtain this value, use the GetLargePageMinimum function.
    # If you specify this value, you must also specify MEM_RESERVE and MEM_COMMIT.
    MEM_LARGE_PAGES = 0x20000000

    # Reserves an address range that can be used to map Address Windowing Extensions (AWE) pages.
    # This value must be used with MEM_RESERVE and no other values.
    MEM_PHYSICAL = 0x00400000

    # Allocates memory at the highest possible address. This can be slower than regular
    # allocations, especially when there are many allocations.
    MEM_TOP_DOWN = 0x00100000

    # Indicates reserved pages where a range of the process's virtual address
    # space is reserved without any physical storage being allocated. For reserved
    # pages, the information in the Protect member is undefined.
    MEM_RESERVE = 0x2000

    # Indicates that data in the memory range is no longer of interest. The pages
    # should not be read from or written to the paging file. However, the memory
    # block will be used again later, so it should not be decommitted. This value
    # cannot be used with any other value.
    MEM_RESET = 0x00080000

    # MEM_RESET_UNDO should only be called on an address range to which MEM_RESET
    # was successfully applied earlier. It indicates that the data in the specified
    # memory range specified by lpAddress and dwSize is of interest to the caller
    # and attempts to reverse the effects of MEM_RESET. If the function succeeds,
    # that means all data in the specified address range is intact. If the function
    # fails, at least some of the data in the address range has been replaced with
    # zeroes. This value cannot be used with any other value.
    MEM_RESET_UNDO = 0x1000000



# -*- coding: utf-8 -*-
from enum import Enum


class MemoryProtectionsEnum(Enum):
    """
    Enum with all protections for a memory page.
    """
    # Enables execute access to the committed region of pages. An attempt to write to the committed
    # region results in an access violation. This flag is not supported by the CreateFileMapping function.
    PAGE_EXECUTE = 0x10

    # Enables execute or read-only access to the committed region of pages. An attempt to write to the committed region
    # results in an access violation. Windows Server 2003 and Windows XP: This attribute is not supported by the
    # CreateFileMapping function until Windows XP with SP2 and Windows Server 2003 with SP1.
    PAGE_EXECUTE_READ = 0x20

    # Enables execute, read-only, or read/write access to the committed region of pages. Windows Server 2003 and
    # Windows XP: This attribute is not supported by the CreateFileMapping function until Windows XP with SP2
    # and Windows Server 2003 with SP1.
    PAGE_EXECUTE_READWRITE = 0x40

    # Enables execute, read-only, or copy-on-write access to a mapped view of a file mapping object. An attempt to
    # write to a committed copy-on-write page results in a private copy of the page being made for the process. The
    # private page is marked as PAGE_EXECUTE_READWRITE, and the change is written to the new page. This flag is not
    # supported by the VirtualAlloc or VirtualAllocEx functions. Windows Vista, Windows Server 2003 and Windows XP:
    # This attribute is not supported by the CreateFileMapping function until Windows Vista with SP1 and Windows Server 2008.
    PAGE_EXECUTE_WRITECOPY = 0x80

    # Pages in the region become guard pages. Any attempt to access a guard page causes the system to raise a
    # STATUS_GUARD_PAGE_VIOLATION exception and turn off the guard page status. Guard pages thus act as a one-time access
    # alarm. For more information, see Creating Guard Pages. When an access attempt leads the system to turn off guard page
    # status, the underlying page protection takes over. If a guard page exception occurs during a system service, the
    # service typically returns a failure status indicator. This value cannot be used with PAGE_NOACCESS. This flag is not
    # supported by the CreateFileMapping function.
    PAGE_GUARD = 0x100

    # Disables all access to the committed region of pages. An attempt to read from, write to, or execute the committed
    # region results in an access violation. This flag is not supported by the CreateFileMapping function.
    PAGE_NOACCESS = 0x01

    # Sets all pages to be non-cachable. Applications should not use this attribute except when explicitly required for a
    # device. Using the interlocked functions with memory that is mapped with SEC_NOCACHE can result in an
    # EXCEPTION_ILLEGAL_INSTRUCTION exception. The PAGE_NOCACHE flag cannot be used with the PAGE_GUARD, PAGE_NOACCESS, or
    # PAGE_WRITECOMBINE flags. The PAGE_NOCACHE flag can be used only when allocating private memory with the VirtualAlloc,
    # VirtualAllocEx, or VirtualAllocExNuma functions. To enable non-cached memory access for shared memory, specify the
    # SEC_NOCACHE flag when calling the CreateFileMapping function.
    PAGE_NOCACHE = 0x200

    # Enables read-only access to the committed region of pages. An attempt to write to the committed region results in
    # an access violation. If Data Execution Prevention is enabled, an attempt to execute code in the committed region
    # results in an access violation.
    PAGE_READONLY = 0x02

    # Enables read-only or read/write access to the committed region of pages. If Data Execution Prevention is enabled,
    # attempting to execute code in the committed region results in an access violation.
    PAGE_READWRITE = 0x04

    # Indicates memory page is readable. (Custom constant)
    PAGE_READABLE = PAGE_EXECUTE_READ | PAGE_EXECUTE_READWRITE | PAGE_READWRITE

    # Indicates memory page is readable and writeable. (Custom constant)
    PAGE_READWRITEABLE = PAGE_EXECUTE_READWRITE | PAGE_READWRITE

    # Sets all locations in the pages as invalid targets for CFG. Used along with any execute page protection like
    # PAGE_EXECUTE, PAGE_EXECUTE_READ, PAGE_EXECUTE_READWRITE and PAGE_EXECUTE_WRITECOPY. Any indirect call to locations
    # in those pages will fail CFG checks and the process will be terminated. The default behavior for executable pages
    # allocated is to be marked valid call targets for CFG. This flag is not supported by the VirtualProtect or
    # CreateFileMapping functions.
    PAGE_TARGETS_INVALID = 0x40000000

    # Pages in the region will not have their CFG information updated while the protection changes for VirtualProtect.
    # For example, if the pages in the region was allocated using PAGE_TARGETS_INVALID, then the invalid information
    # will be maintained while the page protection changes. This flag is only valid when the protection changes to an
    # executable type like PAGE_EXECUTE, PAGE_EXECUTE_READ, PAGE_EXECUTE_READWRITE and PAGE_EXECUTE_WRITECOPY. The default
    # behavior for VirtualProtect protection change to executable is to mark all locations as valid call targets for CFG.
    PAGE_TARGETS_NO_UPDATE = 0x40000000

    # Enables read-only or copy-on-write access to a mapped view of a file mapping object. An attempt to write to a
    # committed copy-on-write page results in a private copy of the page being made for the process. The private page
    # is marked as PAGE_READWRITE, and the change is written to the new page. If Data Execution Prevention is enabled,
    # attempting to execute code in the committed region results in an access violation. This flag is not supported by
    # the VirtualAlloc or VirtualAllocEx functions.
    PAGE_WRITECOPY = 0x08

    # Sets all pages to be write-combined. Applications should not use this attribute except when explicitly required for a
    # device. Using the interlocked functions with memory that is mapped as write-combined can result in an
    # EXCEPTION_ILLEGAL_INSTRUCTION exception. The PAGE_WRITECOMBINE flag cannot be specified with the PAGE_NOACCESS,
    # PAGE_GUARD, and PAGE_NOCACHE flags. The PAGE_WRITECOMBINE flag can be used only when allocating private memory with
    # the VirtualAlloc, VirtualAllocEx, or VirtualAllocExNuma functions. To enable write-combined memory access for shared
    # memory, specify the SEC_WRITECOMBINE flag when calling the CreateFileMapping function. Windows Server 2003 and
    # Windows XP: This flag is not supported until Windows Server 2003 with SP1.
    PAGE_WRITECOMBINE = 0x400



# -*- coding: utf-8 -*-

from ctypes import Structure, WINFUNCTYPE, c_bool, c_ulonglong, c_void_p, sizeof, wintypes


class MEMORY_BASIC_INFORMATION_32(Structure):
    _fields_ = [
        ("BaseAddress", wintypes.DWORD),
        ("AllocationBase", wintypes.DWORD),
        ("AllocationProtect", wintypes.DWORD),
        ("RegionSize", wintypes.DWORD),
        ("State", wintypes.DWORD),
        ("Protect", wintypes.DWORD),
        ("Type", wintypes.DWORD),
    ]


class MEMORY_BASIC_INFORMATION_64(Structure):
    _fields_ = [
        ("BaseAddress", c_ulonglong),
        ("AllocationBase", c_ulonglong),
        ("AllocationProtect", wintypes.DWORD),
        ("__alignment1", wintypes.DWORD),
        ("RegionSize", c_ulonglong),
        ("State", wintypes.DWORD),
        ("Protect", wintypes.DWORD),
        ("Type", wintypes.DWORD),
        ("__alignment2", wintypes.DWORD),
    ]


class SYSTEM_INFO(Structure):
    _fields_ = [
        ("wProcessorArchitecture", wintypes.WORD),
        ("wReserved", wintypes.WORD),
        ("dwPageSize", wintypes.DWORD),
        ("lpMinimumApplicationAddress", c_void_p),
        ("lpMaximumApplicationAddress", c_void_p),
        ("dwActiveProcessorMask", c_void_p),
        ("dwNumberOfProcessors", wintypes.DWORD),
        ("dwProcessorType", wintypes.DWORD),
        ("dwAllocationGranularity", wintypes.DWORD),
        ("wProcessorLevel", wintypes.WORD),
        ("wProcessorRevision", wintypes.WORD),
    ]


# The structure changes according to the Python version (64 or 32 bits).
MEMORY_BASIC_INFORMATION = MEMORY_BASIC_INFORMATION_64 if sizeof(c_void_p) == 8 else MEMORY_BASIC_INFORMATION_32

# For EnumWindows and EnumDesktopWindows functions.
WNDENUMPROC = WINFUNCTYPE(c_bool, wintypes.HWND, wintypes.LPARAM)



# -*- coding: utf-8 -*-

import psutil
import sys

if "win" in sys.platform:
    from ..win32.functions import GetProcessIdByWindowTitle


def get_process_id_by_process_name(process_name: str) -> int:
    """
    Get a process name and return its process ID.
    """
    for process in psutil.process_iter():
        if process.name() == process_name:
            return process.pid


def get_process_id_by_window_title(window_title: str) -> int:
    """
    Get a window title and return its process ID.
    """
    if "win" not in sys.platform:
        raise OSError("This function is compatible only with Windows OS.")

    return GetProcessIdByWindowTitle(window_title)


def pid_exists(pid: int) -> bool:
    """
    Check if the process ID exists.
    """
    return psutil.pid_exists(pid)



# -*- coding: utf-8 -*-
from enum import Enum


class ProcessOperationsEnum(Enum):
    """
    Enum with all permissions and operations you can do to a process.
    """
    # All possible access rights for a process object.Windows Server 2003 and Windows XP: The size of
    # the PROCESS_ALL_ACCESS flag increased on Windows Server 2008 and Windows Vista. If an application
    # compiled for Windows Server 2008 and Windows Vista is run on Windows Server 2003 or Windows XP,
    # the PROCESS_ALL_ACCESS flag is too large and the function specifying this flag fails with
    # ERROR_ACCESS_DENIED. To avoid this problem, specify the minimum set of access rights required for
    # the operation. If PROCESS_ALL_ACCESS must be used, set _WIN32_WINNT to the minimum operating
    # system targeted by your application (for example, #define _WIN32_WINNT _WIN32_WINNT_WINXP). For
    # more information, see Using the Windows Headers.
    PROCESS_ALL_ACCESS = 0x1f0fff

    # Required to create a process.
    PROCESS_CREATE_PROCESS = 0x0080

    # Required to create a thread.
    PROCESS_CREATE_THREAD = 0x0002

    # Required to duplicate a handle using DuplicateHandle.
    PROCESS_DUP_HANDLE = 0x0040

    # Required to retrieve certain information about a process, such as its token, exit code, and priority
    # class (see OpenProcessToken).
    PROCESS_QUERY_INFORMATION = 0x0400

    # Required to retrieve certain information about a process (see GetExitCodeProcess, GetPriorityClass,
    # IsProcessInJob, QueryFullProcessImageName). A handle that has the PROCESS_QUERY_INFORMATION access right
    # is automatically granted PROCESS_QUERY_LIMITED_INFORMATION.Windows Server 2003 and Windows XP: This
    # access right is not supported.
    PROCESS_QUERY_LIMITED_INFORMATION = 0x1000

    # Required to set certain information about a process, such as its priority class (see SetPriorityClass).
    PROCESS_SET_INFORMATION = 0x0200
    PROCESS_SET_LIMITED_INFORMATION = 0x2000

    # Required to set memory limits using SetProcessWorkingSetSize.
    PROCESS_SET_QUOTA = 0x0100

    # Required to suspend or resume a process.
    PROCESS_SUSPEND_RESUME = 0x0800

    # Required to terminate a process using TerminateProcess.
    PROCESS_TERMINATE = 0x0800

    # Required to perform an operation on the address space of a process (see VirtualProtectEx and WriteProcessMemory).
    PROCESS_VM_OPERATION = 0x0008

    # Required to read memory in a process using ReadProcessMemory.
    PROCESS_VM_READ = 0x0010

    # Required to write to memory in a process using WriteProcessMemory.
    PROCESS_VM_WRITE = 0x0020



# -*- coding: utf-8 -*-

from tkinter import DoubleVar, Frame, Label, Menu, Listbox, Scrollbar, Tk, filedialog
from tkinter.ttk import Button, Entry, Menubutton, Progressbar
from typing import Tuple, Type, TypeVar, Union

from PyMemoryEditor import ScanTypesEnum
from PyMemoryEditor.process import AbstractProcess

import json


T = TypeVar("T")


class ApplicationWindow(Tk):
    """
    Main window of the application.
    """
    __comparison_methods = {
        ScanTypesEnum.EXACT_VALUE: lambda x, y: x == y,
        ScanTypesEnum.NOT_EXACT_VALUE: lambda x, y: x != y,
        ScanTypesEnum.BIGGER_THAN: lambda x, y: x > y,
        ScanTypesEnum.SMALLER_THAN: lambda x, y: x < y,
        ScanTypesEnum.VALUE_BETWEEN: lambda x, y: y[0] <= x <= y[1],
        ScanTypesEnum.NOT_VALUE_BETWEEN: lambda x, y: y[0] > x or x > y[1],
    }

    __max_listbox_length = 200

    def __init__(self, process: AbstractProcess):
        super().__init__()
        self.__process = process

        self.__scan_type = ScanTypesEnum.EXACT_VALUE
        self.__value_type = int
        self.__value_length = 4

        self.__addresses = dict()
        self.__selected_page = 0
        self.__max_page = 0

        self.__finding_addresses = False  # Indicate it is searching for addresses (first step of a new scan).
        self.__scanning = False           # Indicate a scan has started.
        self.__updating = False           # Indicate it is updating the values of the found addresses.

        self["bg"] = "white"

        self.title(f"PyMemoryEditor (Sample) - Process ID: {process.pid}")
        self.geometry("1100x400")
        self.resizable(False, False)

        self.protocol("WM_DELETE_WINDOW", self.__on_close)
        self.__close = False

        self.__build()
        self.mainloop()

    def __build(self) -> None:
        """
        Build the widgets of the window.
        """
        # Register to validate numeric entries.
        self.__entry_register_int = self.register(self.__validate_int_entry)
        self.__entry_register_hex = self.register(self.__validate_hex_entry)

        # Frame for scan input.
        self.__input_frame_1 = Frame(self)
        self.__input_frame_1["bg"] = "white"
        self.__input_frame_1.pack(padx=5, fill="x", expand=True)

        self.__scan_input_frame = Frame(self.__input_frame_1)
        self.__scan_input_frame["bg"] = "white"
        self.__scan_input_frame.pack(fill="x", expand=True)

        # Value input.
        self.__values_frame = Frame(self.__scan_input_frame)
        self.__values_frame["bg"] = "white"
        self.__values_frame.pack(side="left", fill="x", expand=True)

        self.__value_label = Label(self.__values_frame, text="Value: ", bg="white", font=("Arial", 12))
        self.__value_label.pack(side="left")

        self.__value_entry = Entry(self.__values_frame)
        self.__value_entry.pack(side="left", expand=True, fill="x")

        self.__second_value_entry = Entry(self.__values_frame)

        Label(self.__scan_input_frame, bg="white").pack(side="left")

        # Value length.
        Label(self.__scan_input_frame, text="Length (Bytes): ", bg="white", font=("Arial", 12)).pack(side="left")

        self.__length_entry = Entry(self.__scan_input_frame, width=5)
        self.__length_entry.insert(0, "4")
        self.__length_entry.config(validate="key", validatecommand=(self.__entry_register_int, "%P"))
        self.__length_entry.pack(side="left")

        Label(self.__scan_input_frame, bg="white").pack(side="left", padx=5)

        # Value type input.
        self.__type_menu_button = Menubutton(self.__scan_input_frame, width=10)
        self.__type_menu_button.pack(side="left")

        self.__type_menu = Menu(tearoff=0, bg="white")
        self.__type_menu.add_command(label="Boolean", command=lambda: self.__set_value_type(0))
        self.__type_menu.add_command(label="Integer", command=lambda: self.__set_value_type(1))
        self.__type_menu.add_command(label="Float", command=lambda: self.__set_value_type(2))
        self.__type_menu.add_command(label="String", command=lambda: self.__set_value_type(3))
        self.__type_menu_button.config(menu=self.__type_menu, text="Integer")

        Label(self.__scan_input_frame, bg="white").pack(side="left", padx=10)

        # Scan type input.
        Label(self.__scan_input_frame, text="Scan Type: ", bg="white", font=("Arial", 12)).pack(side="left")

        self.__scan_menu_button = Menubutton(self.__scan_input_frame, width=20)
        self.__scan_menu_button.pack(side="left")

        self.__scan_menu = Menu(tearoff=0, bg="white")
        self.__scan_menu.add_command(label="Exact Value", command=lambda: self.__set_scan_type(0))
        self.__scan_menu.add_command(label="Not Exact Value", command=lambda: self.__set_scan_type(1))
        self.__scan_menu.add_command(label="Smaller Than", command=lambda: self.__set_scan_type(2))
        self.__scan_menu.add_command(label="Bigger Than", command=lambda: self.__set_scan_type(3))
        self.__scan_menu.add_command(label="Value Between", command=lambda: self.__set_scan_type(4))
        self.__scan_menu.add_command(label="Not Value Between", command=lambda: self.__set_scan_type(5))
        self.__scan_menu_button.config(menu=self.__scan_menu, text="Exact Value")

        Label(self.__scan_input_frame, bg="white").pack(side="left", padx=5)

        # Buttons for scanning.
        self.__new_scan_button = Button(self.__scan_input_frame, text="First Scan", command=self.__new_scan)
        self.__new_scan_button.pack(side="left")

        Label(self.__scan_input_frame, bg="white").pack(side="left")

        self.__next_scan_button = Button(self.__scan_input_frame, command=self.__next_scan)
        self.__next_scan_button.pack(side="left")

        # Progress bar for scanning and updating.
        self.__progress_var = DoubleVar()

        self.__progress_bar = Progressbar(self.__input_frame_1, variable=self.__progress_var)
        self.__progress_bar.pack(pady=5, fill="x", expand=True)

        # Label for counting and buttons for changing page and updating values.
        self.__result_frame = Frame(self)
        self.__result_frame["bg"] = "white"
        self.__result_frame.pack(padx=5, fill="both", expand=True)

        self.__count_frame = Frame(self.__result_frame)
        self.__count_frame["bg"] = "white"
        self.__count_frame.pack(pady=5, fill="x", expand=True)

        self.__count_label = Label(self.__count_frame, font=("Arial", 8), bg="white")
        self.__count_label.config(text="Start a new scan to find memory addresses.")
        self.__count_label.pack(side="left")

        Button(self.__count_frame, text="Update Values", command=self.__update_values).pack(side="right")
        Label(self.__count_frame, bg="white").pack(side="right", padx=10)

        Button(self.__count_frame, text="Next Page", command=lambda: self.__change_results_page(1)).pack(side="right")

        self.__page_label = Label(self.__count_frame, text="0 of 0", width=12, borderwidth=2, relief="solid")
        self.__page_label.pack(side="right", padx=10)

        Button(self.__count_frame, text="Previous Page", command=lambda: self.__change_results_page(-1)).pack(side="right")

        # List with addresses and their values.
        self.__list_frame = Frame(self.__result_frame)
        self.__list_frame["bg"] = "white"
        self.__list_frame.pack(fill="both", expand=True)

        self.__scrollbar = Scrollbar(self.__list_frame, orient="vertical", command=self.__on_move_list_box)

        self.__address_list = Listbox(self.__list_frame, width=20)
        self.__address_list.bind("<MouseWheel>", self.__on_mouse_wheel)
        self.__address_list.bind("<<ListboxSelect>>", self.__select_address)
        self.__address_list.config(yscrollcommand=self.__scrollbar.set)
        self.__address_list.pack(side="left", fill="y")

        self.__value_list = Listbox(self.__list_frame)
        self.__value_list.bind("<MouseWheel>", self.__on_mouse_wheel)
        self.__value_list.bind("<<ListboxSelect>>", self.__select_value)
        self.__value_list.config(yscrollcommand=self.__scrollbar.set)
        self.__value_list.pack(side="left", fill="both", expand=True)

        self.__scrollbar.pack(side="left", fill="y")

        # Frame and widgets to allow user changing the value of a memory address.
        self.__input_frame_2 = Frame(self)
        self.__input_frame_2["bg"] = "white"
        self.__input_frame_2.pack(padx=5, fill="x", expand=True)

        Label(self.__input_frame_2, text="Address:", bg="white").pack(side="left")

        self.__address_entry = Entry(self.__input_frame_2)
        self.__address_entry.config(validate="key", validatecommand=(self.__entry_register_hex, "%P"))
        self.__address_entry.pack(side="left")

        Label(self.__input_frame_2, bg="white").pack(side="left")

        Label(self.__input_frame_2, text="New Value:", bg="white").pack(side="left")

        self.__new_value_entry = Entry(self.__input_frame_2)
        self.__new_value_entry.pack(side="left", fill="x", expand=True)

        Button(self.__input_frame_2, text="Replace", command=self.__write_value).pack(side="left")

        Label(self.__input_frame_2, bg="white").pack(side="left")

        Button(self.__input_frame_2, text="Export Data", command=self.__export_data).pack(side="left")

    def __change_results_page(self, step: int):
        """
        Change the page of results.
        """
        if step != 0 and (self.__finding_addresses or self.__updating): return

        max_page = len(self.__addresses) // self.__max_listbox_length

        if self.__selected_page > max_page:
            self.__selected_page = max_page

        next_page = self.__selected_page + step

        if next_page < 0 or next_page > max_page: return

        if not (0 <= next_page <= max_page):
            next_page = self.__selected_page

        text = f"{next_page} of {max_page}"
        self.__page_label.config(text=text)

        self.__selected_page = next_page
        self.__update_listboxes()

    def __check_address_entry(self, address: str) -> bool:
        """
        Check if the address entry is valid.
        """
        try:
            if int(address, 16) in self.__addresses:
                return True
            raise ValueError()
        except ValueError:
            self.__address_entry.delete(0, "end")
            self.__address_entry.insert(0, "00000000")
        return False

    def __check_value_entry(self, value: str, value_type: Type, length: int, entry: Entry) -> bool:
        """
        Check if the new value entry is valid.
        """
        if length == 0:
            self.__length_entry.delete(0, "end")
            self.__length_entry.insert(0, "1")
            return False

        try:
            if value and str(value_type(value)) == value and (value_type is not str or len(value) <= length):
                return True
            raise ValueError()

        except ValueError:
            entry.delete(0, "end")
            entry.insert(0, "Invalid value")
        return False

    def __export_data(self):
        """
        Export found addresses and values from the scan.
        """
        data = self.__addresses.copy()

        filename = filedialog.asksaveasfilename(
            title="Save as...",
            filetypes=(
                ("JSON (*.json)", "*.json"),
                ("All files (*.*)", "*.*"),
            ),
            defaultextension=".json"
        )
        if not filename: return

        with open(filename, "w") as file:
            data = json.dumps(data, indent=4)
            file.write(data)

    def __new_scan(self) -> None:
        """
        Start a new seach at the whole memory of the process.
        """
        if self.__finding_addresses or self.__updating: return

        # If a scan is already in progress, clear all results for a new scan.
        if self.__scanning: return self.__stop_scan()

        # Get the inputs.
        value = self.__value_entry.get().strip()
        value_2 = self.__second_value_entry.get().strip()

        length = int(self.__length_entry.get())
        pytype = self.__value_type
        scan_type = self.__scan_type

        # Validate the input.
        if not self.__check_value_entry(value, pytype, length, self.__value_entry): return

        value = pytype(value)

        if scan_type in [ScanTypesEnum.VALUE_BETWEEN, ScanTypesEnum.NOT_VALUE_BETWEEN]:
            if not self.__check_value_entry(value_2, pytype, length, self.__second_value_entry): return
            value = (value, pytype(value_2))

        # Start the scan.
        self.__value_length = length

        self.after(100, lambda: self.__start_scan(pytype, length, value, scan_type))

    def __next_scan(self) -> None:
        """
        Filter the found addresses.
        """
        self.__update_values(remove=True)

    def __on_close(self, *args) -> None:
        """
        Event to close the program graciously.
        """
        self.__close = True
        self.update()

        if self.__updating or self.__finding_addresses:
            self.after(10, self.__on_close)
            return

        self.destroy()

    def __on_mouse_wheel(self, event) -> str:
        """
        Event to sync the listboxes.
        """
        self.__address_list.yview("scroll", event.delta, "units")
        self.__value_list.yview("scroll", event.delta, "units")
        return "break"

    def __on_move_list_box(self, *args) -> None:
        """
        Event to sync the listboxes.
        """
        self.__address_list.yview(*args)
        self.__value_list.yview(*args)

    def __select_address(self, event) -> None:
        """
        Event to get the selected address and copy it.
        """
        selection = event.widget.curselection()
        if not selection: return

        address = self.__address_list.get(int(selection[0])).split(" ")[-1]
        if not address: return

        self.__address_entry.delete(0, "end")
        self.__address_entry.insert(0, address)

    def __select_value(self, event) -> None:
        """
        Event to get the selected value and copy it.
        """
        selection = event.widget.curselection()
        if not selection: return

        value = self.__value_list.get(int(selection[0]))[len("Value: "):]
        self.__new_value_entry.delete(0, "end")
        self.__new_value_entry.insert(0, value)

    def __set_scan_type(self, scan_type: int) -> None:
        """
        Method for the Menubutton to select a scan type.
        """
        # Allow select a new scan type only if program is not getting new addresses or updating their values.
        if self.__finding_addresses or self.__updating: return

        self.__scan_type = [
            ScanTypesEnum.EXACT_VALUE,
            ScanTypesEnum.NOT_EXACT_VALUE,
            ScanTypesEnum.SMALLER_THAN,
            ScanTypesEnum.BIGGER_THAN,
            ScanTypesEnum.VALUE_BETWEEN,
            ScanTypesEnum.NOT_VALUE_BETWEEN
        ][scan_type]

        if self.__scan_type in [ScanTypesEnum.VALUE_BETWEEN, ScanTypesEnum.NOT_VALUE_BETWEEN]:
            self.__value_label.config(text="Values:")
            self.__second_value_entry.pack(padx=5, side="left", expand=True, fill="x")
        else:
            self.__value_label.config(text="Value:")
            self.__second_value_entry.delete(0, "end")
            self.__second_value_entry.forget()

        text = " ".join(word.capitalize() for word in self.__scan_type.name.split("_"))
        self.__scan_menu_button.config(text=text)

    def __set_value_type(self, value_type: int):
        """
        Method for the Menubutton to select a value type.
        """
        if self.__scanning: return

        self.__value_type = [bool, int, float, str][value_type]
        self.__type_menu_button.config(text=["Boolean", "Integer", "Float", "String"][value_type])

    def __start_scan(self, pytype: Type[T], length: int, value: Union[T, Tuple[T, T]], scan_type: ScanTypesEnum) -> None:
        """
        Search for a value on the whole memory of the process.
        """
        self.__new_scan_button.config(text="Scanning")
        self.__count_label.config(text=f"Found {len(self.__addresses)} addresses.")
        self.update()

        self.__finding_addresses = True
        self.__scanning = True

        # Get a generator object to find the addresses by a value or within a range.
        if scan_type in [ScanTypesEnum.VALUE_BETWEEN, ScanTypesEnum.NOT_VALUE_BETWEEN]:
            address_finder = self.__process.search_by_value_between(
                pytype, length, value[0], value[1], progress_information=True,
                not_between=scan_type is ScanTypesEnum.NOT_VALUE_BETWEEN,
            )
        else:
            address_finder = self.__process.search_by_value(pytype, length, value, scan_type, progress_information=True)

        # Search for the addresses and add the results to the listbox.
        for address, info in address_finder:
            if self.__close: break

            self.__progress_var.set(info["progress"] * 100)
            self.__addresses[address] = "loading..."
            self.update()

            self.__count_label.config(text=f"Found {len(self.__addresses)} addresses.")

        # Get the value of each address and update the listbox.
        self.__finding_addresses = False
        self.__update_values()

        self.__new_scan_button.config(text="New Scan")
        self.__next_scan_button.config(text="Next Scan")
        self.__progress_var.set(100)

    def __stop_scan(self) -> None:
        """
        Clear all results and get everything ready for a new scan.
        """
        self.__count_label.config(text="Start a new scan to find memory addresses.")
        self.__new_scan_button.config(text="First Scan")
        self.__next_scan_button.config(text="")

        self.__address_list.delete(0, "end")
        self.__value_list.delete(0, "end")

        self.__progress_var.set(0)

        self.__scanning = False
        self.__addresses = dict()

        self.__change_results_page(0)

    def __validate_int_entry(self, string: str) -> bool:
        """
        Method to validate if an input is integer.
        """
        if self.__scanning: return False

        for char in string:
            if char not in "0123456789": return False
        return True

    def __validate_hex_entry(self, string: str) -> bool:
        """
        Method to validate if an input is hexadecimal.
        """
        for char in string.upper():
            if char not in "0123456789ABCDEF": return False
        return True

    def __update_listboxes(self) -> None:
        """
        Update the listboxes with the found addresses and theirs values.
        """
        start = self.__selected_page * self.__max_listbox_length

        items = [(address, value) for address, value in self.__addresses.items()]
        items = items[start: start + self.__max_listbox_length]

        self.__address_list.delete(0, "end")
        self.__value_list.delete(0, "end")

        for address, value in items:
            self.__address_list.insert("end", f"Addr: {hex(address)[2:].upper()}")
            self.__value_list.insert("end", f"Value: {value}")
            self.update()

    def __update_values(self, *, remove: bool = False) -> None:
        """
        Update the values of the found addresses. If "remove" is True, it will
        compare the current value in memory and remove the address from the
        results if the comparison is False.
        """
        if self.__updating or self.__finding_addresses: return
        if not self.__addresses: return self.__progress_var.set(100)

        # Get the value to compare.
        expected_value = self.__value_entry.get().strip()
        expected_value_2 = self.__second_value_entry.get().strip()

        value_type = self.__value_type
        value_length = self.__value_length

        if not self.__check_value_entry(expected_value, value_type, value_length, self.__value_entry): return
        expected_value = value_type(expected_value)

        if self.__scan_type in [ScanTypesEnum.VALUE_BETWEEN, ScanTypesEnum.NOT_VALUE_BETWEEN]:
            if not self.__check_value_entry(expected_value_2, value_type, value_length, self.__second_value_entry): return
            expected_value = (expected_value, value_type(expected_value_2))

        # Get the comparison method.
        compare = self.__comparison_methods[self.__scan_type]

        # Indicate the application is updating the values.
        self.__updating = True
        self.__progress_var.set(0)

        # Tell user application is updating the values.
        new_scan_button_text = self.__new_scan_button["text"]
        self.__new_scan_button.config(text="Updating")

        # Get the address and its current value in memory.
        total, count, index = len(self.__addresses), 0, 0

        for address, current_value in self.__process.search_by_addresses(value_type, value_length, self.__addresses):
            self.__progress_var.set((count / total) * 100)
            self.update()

            count += 1

            # Return if user asked for closing the application.
            if self.__close:
                self.__updating = False
                return

            # If value is corrupted or "remove" is True and comparison is False, remove the value from the results.
            if current_value is None or (remove and not compare(current_value, expected_value)):
                self.__address_list.delete(index)
                self.__value_list.delete(index)
                self.__addresses.pop(address)

            else:
                self.__addresses[address] = current_value
                index += 1

        # Start the process of updating the listboxes.
        self.__change_results_page(0)
        self.__update_listboxes()

        # Indicate update has finished.
        self.__new_scan_button.config(text=new_scan_button_text)
        self.__updating = False

        self.__count_label.config(text=f"Found {len(self.__addresses)} addresses.")
        self.__progress_var.set(100)

    def __write_value(self) -> None:
        """
        Change the value in memory of an address of the result list.
        """
        address = self.__address_entry.get().strip()
        if not self.__check_address_entry(address): return

        # Get the inputs.
        address = int(address, 16)
        value = self.__new_value_entry.get()
        pytype = self.__value_type
        length = self.__value_length

        # Validate the input.
        if not self.__check_value_entry(value, pytype, length, self.__new_value_entry): return

        # Write the new value.
        self.__process.write_process_memory(address, pytype, length, pytype(value))



# -*- coding: utf-8 -*-
from PyMemoryEditor import __version__

from .main_application_window import ApplicationWindow
from .open_process_window import OpenProcessWindow

import sys


def main(*args, **kwargs):
    if len(sys.argv) > 1 and sys.argv[1].strip() in ["--version", "-v"]:
        return print(__version__)

    open_process_window = OpenProcessWindow()
    process = open_process_window.get_process()

    if not process: return

    try: ApplicationWindow(process)
    finally: process.close()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from ..enums import ScanTypesEnum
from ..process import AbstractProcess
from ..process.errors import ClosedProcess
from .functions import (
    get_memory_regions,
    read_process_memory,
    search_addresses_by_value,
    search_values_by_addresses,
    write_process_memory
)
from typing import Generator, Optional, Sequence, Tuple, Type, TypeVar, Union


T = TypeVar("T")


class LinuxProcess(AbstractProcess):
    """
    Class to open a Linux process for reading, writing and searching at its memory.
    """

    def __init__(
        self,
        *,
        window_title: Optional[str] = None,
        process_name: Optional[str] = None,
        pid: Optional[int] = None,
        **kwargs
    ):
        """
        :param window_title: window title of the target program.
        :param process_name: name of the target process.
        :param pid: process ID.
        """
        super().__init__(
            window_title=window_title,
            process_name=process_name,
            pid=pid
        )
        self.__closed = False

    def close(self) -> bool:
        # Check the documentation of this method in the AbstractProcess superclass for more information.
        self.__closed = True
        return True

    def get_memory_regions(self) -> Generator[dict, None, None]:
        # Check the documentation of this method in the AbstractProcess superclass for more information.
        if self.__closed: raise ClosedProcess()
        return get_memory_regions(self.pid)

    def read_process_memory(
        self,
        address: int,
        pytype: Type[T],
        bufflength: int
    ) -> T:
        # Check the documentation of this method in the AbstractProcess superclass for more information.
        if self.__closed: raise ClosedProcess()
        return read_process_memory(self.pid, address, pytype, bufflength)

    def search_by_addresses(
        self,
        pytype: Type[T],
        bufflength: int,
        addresses: Sequence[int],
        *,
        raise_error: bool = False,
    ) -> Generator[Tuple[int, Optional[T]], None, None]:

        # Check the documentation of this method in the AbstractProcess superclass for more information.
        if self.__closed: raise ClosedProcess()
        return search_values_by_addresses(self.pid, pytype, bufflength, addresses, raise_error=raise_error)

    def search_by_value(
        self,
        pytype: Type[T],
        bufflength: int,
        value: Union[bool, int, float, str, bytes],
        scan_type: ScanTypesEnum = ScanTypesEnum.EXACT_VALUE,
        *,
        progress_information: bool = False,
        writeable_only: bool = False,
    ) -> Generator[Union[int, Tuple[int, dict]], None, None]:

        # Check the documentation of this method in the AbstractProcess superclass for more information.
        if self.__closed: raise ClosedProcess()

        if scan_type in [ScanTypesEnum.VALUE_BETWEEN, ScanTypesEnum.NOT_VALUE_BETWEEN]:
            raise ValueError("Use the method search_by_value_between(...) to search within a range of values.")

        return search_addresses_by_value(self.pid, pytype, bufflength, value, scan_type, progress_information, writeable_only)

    def search_by_value_between(
        self,
        pytype: Type[T],
        bufflength: int,
        start: Union[bool, int, float, str, bytes],
        end: Union[bool, int, float, str, bytes],
        *,
        not_between: bool = False,
        progress_information: bool = False,
        writeable_only: bool = False,
    ) -> Generator[Union[int, Tuple[int, dict]], None, None]:

        # Check the documentation of this method in the AbstractProcess superclass for more information.
        if self.__closed: raise ClosedProcess()

        scan_type = ScanTypesEnum.NOT_VALUE_BETWEEN if not_between else ScanTypesEnum.VALUE_BETWEEN
        return search_addresses_by_value(self.pid, pytype, bufflength, (start, end), scan_type, progress_information, writeable_only)

    def write_process_memory(
        self,
        address: int,
        pytype: Type[T],
        bufflength: int,
        value: Union[bool, int, float, str, bytes]
    ) -> T:
        # Check the documentation of this method in the AbstractProcess superclass for more information.
        if self.__closed: raise ClosedProcess()
        return write_process_memory(self.pid, address, pytype, bufflength, value)



# -*- coding: utf-8 -*-

"""
Multi-platform library developed with ctypes for reading, writing and
searching at process memory, in a simple and friendly way with Python 3.

The package supports Windows and Linux (32-bit and 64-bit).
"""

__author__ = "Jean Loui Bernard Silva de Jesus"
__version__ = "1.5.20"

from .enums import ScanTypesEnum
from .process.errors import ClosedProcess, ProcessIDNotExistsError, ProcessNotFoundError
import sys

# For Windows.
if "win" in sys.platform:
    from .win32.process import WindowsProcess
    from .win32.enums.process_operations import ProcessOperationsEnum
    OpenProcess = WindowsProcess

# For Linux.
else:
    from .linux.process import LinuxProcess
    from .linux.ptrace import ptrace
    from .linux.ptrace.enums import PtraceCommandsEnum
    OpenProcess = LinuxProcess



# -*- coding: utf-8 -*-
from .abstract import AbstractSearchAlgorithm
from typing import Generator, Optional, Sequence


class KMPSearch(AbstractSearchAlgorithm):
    """
    Algorithm Knuth-Morris-Pratt (KMP) for matching pattern in sequences.
    """
    def __init__(self, pattern: Sequence, pattern_length: Optional[int] = None):
        if pattern_length is None:
            pattern_length = len(pattern)

        # Instantiate the parameters.
        self.__pattern = pattern
        self.__pattern_length = pattern_length

        self.__lps: list = [0]  # List to save the LPS (longest prefix which is also a suffix).

        # Process the pattern.
        for index in range(1, self.__pattern_length):
            j = self.__lps[index - 1]

            while j > 0 and pattern[j] != pattern[index]:
                j = self.__lps[j - 1]

            self.__lps.append(j + 1 if pattern[j] == pattern[index] else j)

    def search(self, sequence: Sequence, length: Optional[int] = None) -> Generator[int, None, None]:
        """
        Return all the matching position of pattern.
        """
        if length is None:
            length = len(sequence)

        offset = 0

        for index in range(length):
            while offset > 0 and sequence[index] != self.__pattern[offset]:
                offset = self.__lps[offset - 1]

            if sequence[index] == self.__pattern[offset]:
                offset += 1

            if offset == self.__pattern_length:
                yield index - (offset - 1)
                offset = self.__lps[offset - 1]



# -*- coding: utf-8 -*-
from enum import Enum


class ScanTypesEnum(Enum):
    """
    Enum with scan types.
    """
    EXACT_VALUE = 0
    NOT_EXACT_VALUE = 1
    BIGGER_THAN = 2
    SMALLER_THAN = 3
    BIGGER_THAN_OR_EXACT_VALUE = 4
    SMALLER_THAN_OR_EXACT_VALUE = 5
    VALUE_BETWEEN = 6
    NOT_VALUE_BETWEEN = 7



# -*- coding: utf-8 -*-

class ClosedProcess(Exception):
    def __str__(self):
        return "Operation not allowed on a closed process."


class ProcessIDNotExistsError(Exception):

    def __init__(self, pid: int):
        self.__pid = pid

    def __str__(self) -> str:
        return "The process ID \"%i\" does not exist." % self.__pid


class ProcessNotFoundError(Exception):

    def __init__(self, process_name: str):
        self.__process_name = process_name

    def __str__(self) -> str:
        return "Could not find the process \"%s\"." % self.__process_name


class WindowNotFoundError(Exception):

    def __init__(self, window_title: str):
        self.__window_title = window_title

    def __str__(self) -> str:
        return "Could not find the window \"%s\"." % self.__window_title



# -*- coding: utf-8 -*-
from enum import Enum


class StandardAccessRightsEnum(Enum):
    """
    Enum with of standard access rights that correspond to operations
    common to most types of securable objects.
    """
    # Required to delete the object.
    DELETE = 0x00010000

    # Required to read information in the security descriptor for the object, not including the
    # information in the SACL. To read or write the SACL, you must request the ACCESS_SYSTEM_SECURITY
    # access right. For more information, see SACL Access Right.
    READ_CONTROL = 0x00020000

    # The right to use the object for synchronization. This enables a thread to wait until the object
    # is in the signaled state.
    SYNCHRONIZE = 0x00100000

    # Required to modify the DACL in the security descriptor for the object.
    WRITE_DAC = 0x00040000

    # Required to change the owner in the security descriptor for the object.
    WRITE_OWNER = 0x00080000



# -*- coding: utf-8 -*-

from .errors import ProcessIDNotExistsError, ProcessNotFoundError, WindowNotFoundError
from .util import get_process_id_by_process_name, get_process_id_by_window_title, pid_exists


class ProcessInfo(object):
    """
    Class to save information of a process.
    """

    __pid = 0
    __process_name = ""
    __window_title = ""

    @property
    def pid(self) -> int:
        return self.__pid

    @pid.setter
    def pid(self, pid: int) -> None:

        # Check if the value is an integer.
        if not isinstance(pid, int):
            raise ValueError("The process ID must be an integer.")

        # Check if the PID exists and instantiate it.
        if pid_exists(pid): self.__pid = pid
        else: raise ProcessIDNotExistsError(pid)

    @property
    def process_name(self) -> str:
        return self.__process_name

    @process_name.setter
    def process_name(self, process_name: str) -> None:

        # Get the process ID.
        pid = get_process_id_by_process_name(process_name)
        if not pid: raise ProcessNotFoundError(process_name)

        # Set the PID and process name.
        self.__pid = pid
        self.__process_name = process_name

    @property
    def window_title(self) -> str:
        return self.__window_title

    @window_title.setter
    def window_title(self, window_title: str) -> None:

        # Get the process ID.
        pid = get_process_id_by_window_title(window_title)
        if not pid: raise WindowNotFoundError(window_title)

        # Set the PID and the window title.
        self.__pid = pid
        self.__window_title = window_title



# -*- coding: utf-8 -*-

from .memory_allocation_states import MemoryAllocationStatesEnum
from .memory_protections import MemoryProtectionsEnum
from .memory_types import MemoryTypesEnum
from .process_operations import ProcessOperationsEnum
from .standard_access_rights import StandardAccessRightsEnum



from abc import ABC, abstractmethod
from typing import Generator, Optional, Sequence


class AbstractSearchAlgorithm(ABC):
    @abstractmethod
    def __init__(self, pattern: Sequence, pattern_length: Optional[int] = None):
        raise NotImplementedError()

    @abstractmethod
    def search(self, sequence: Sequence, length: Optional[int] = None) -> Generator[int, None, None]:
        raise NotImplementedError()



# -*- coding: utf-8 -*-

import os
import sys

current_dir = os.getcwd()
sys.path.append(current_dir)

from PyMemoryEditor import OpenProcess, ScanTypesEnum
from PyMemoryEditor import __version__ as version



# -*- coding: utf-8 -*-

from tkinter import Frame, Label, Listbox, Scrollbar, Tk
from tkinter.ttk import Button, Entry, Style
from typing import Optional

from PyMemoryEditor import OpenProcess
from PyMemoryEditor.errors import ProcessIDNotExistsError, ProcessNotFoundError
from PyMemoryEditor.process import AbstractProcess

import psutil


class OpenProcessWindow(Tk):
    """
    Window for opening a process.
    """
    def __init__(self):
        super().__init__()
        self.__process = None

        self["bg"] = "white"

        self.title("PyMemoryEditor (Sample) - Select a process to scan")
        self.geometry("450x350")
        self.resizable(False, False)

        Label(self, text="Select a process or insert the PID or the process name", bg="white", font=("Arial", 10)).pack(padx=20, pady=5)

        self.__list_frame = Frame(self)
        self.__list_frame["bg"] = "white"
        self.__list_frame.pack(padx=38, fill="both", expand=True)

        self.__scrollbar = Scrollbar(self.__list_frame, orient="vertical", command=self.__on_move_list_box)

        self.__process_list = Listbox(self.__list_frame, width=40, borderwidth=1, relief="solid")
        self.__process_list.bind("<<ListboxSelect>>", self.__select_process)
        self.__process_list.config(yscrollcommand=self.__scrollbar.set)
        self.__process_list.pack(side="left", fill="both", expand=True)

        self.__scrollbar.pack(side="left", fill="y")

        self.__input_frame = Frame(self)
        self.__input_frame["bg"] = "white"
        self.__input_frame.pack(padx=38, fill="x", expand=True)

        Label(
            self.__input_frame, text="Process:", bg="#eee",
            borderwidth=1, relief="solid", font=("Arial", 9)
        ).pack(ipadx=3, ipady=1, side="left")

        self.__entry = Entry(self.__input_frame)
        self.__entry.pack(side="left", fill="x", expand=True)

        self.__button_style = Style()
        self.__button_style.configure("TButton", font=('Helvetica', 12))

        Button(self, text="Scan Process", command=self.__open_process, style="TButton").pack(ipadx=5, ipady=5)
        Label(self, bg="white").pack()

        self.__update_process_list()
        self.mainloop()

    def __on_move_list_box(self, *args) -> None:
        """
        Event to sync the listbox.
        """
        self.__process_list.yview(*args)

    def __open_process(self) -> None:
        """
        Open the process by the user input.
        """
        entry = self.__entry.get().strip()

        try:
            self.__process = OpenProcess(pid=int(entry))
            return self.destroy()

        except ValueError:
            try:
                self.__process = OpenProcess(process_name=entry)
                return self.destroy()
            except (ProcessIDNotExistsError, ProcessNotFoundError): pass
        except (ProcessIDNotExistsError, ProcessNotFoundError): pass

        self.__entry.delete(0, "end")
        self.__entry.insert(0, "Process not found.")

    def __select_process(self, event) -> None:
        """
        Event to get the selected address and copy it.
        """
        selection = event.widget.curselection()
        if not selection: return

        index = int(selection[0])
        if index == 0: return self.__process_list.select_clear(0, "end")

        process = int(self.__process_list.get(index).split("-")[0].strip())
        if not process: return

        self.__entry.delete(0, "end")
        self.__entry.insert(0, str(process))

    def __update_process_list(self):
        """
        Update the process list with new processes.
        """
        self.__process_list.delete(0, "end")

        processes = sorted([
            (process.name(), process.pid, process.memory_info().vms) for process in psutil.process_iter()
        ], key=lambda x: x[0].lower())

        self.__process_list.insert("end", "{:<14} {:<17} {}".format("PID", "VMS", "Process Name"))
        self.__process_list.itemconfig(0, {"bg": "#ccc"})

        index = 0

        for name, pid, memory in processes:
            if not name.replace(" ", ""): continue
            name = name[:-3] + "..." if len(name) > 35 else name

            self.__process_list.insert("end", "{:0>7} - {:0>7} KB - {}".format(pid, memory // 1024, name))
            self.__process_list.itemconfig(index + 1, {"bg": ["white", "#ddd"][index % 2]})

            index += 1

    def get_process(self) -> Optional[AbstractProcess]:
        """
        Return the opened process.
        """
        return self.__process



# -*- coding: utf-8 -*-

# Read more about operations with processes by ptrace system call here:
# https://man7.org/linux/man-pages/man2/ptrace.2.html
# https://refspecs.linuxbase.org/LSB_5.0.0/LSB-Core-generic/LSB-Core-generic/baselib-ptrace-1.html
# ...

from .enums import PtraceCommandsEnum

from ctypes.util import find_library
import ctypes

libc = ctypes.CDLL(find_library("c"), use_errno=True)
libc.ptrace.argtypes = (ctypes.c_ulong,) * 4
libc.ptrace.restype = ctypes.c_long


def ptrace(command: PtraceCommandsEnum, pid: int, *args: int) -> int:
    """
    Run ptrace() system call with the provided command, pid and arguments.
    """
    result = libc.ptrace(command.value, pid, *args)

    if result == -1:
        error_no = ctypes.get_errno()

        if error_no:
            error_msg = ctypes.string_at(libc.strerror(error_no))
            raise OSError(error_msg)

    return result



# -*- coding: utf-8 -*-
from .abstract import AbstractSearchAlgorithm
from typing import Generator, Optional, Sequence, Union


class BMHSearch(AbstractSearchAlgorithm):
    """
    Algorithm Boyer-Moore-Horspool (BMH) for matching pattern in sequences.
    """
    def __init__(self, pattern: Sequence, pattern_length: Optional[int] = None, alphabet_length: int = 256):
        if pattern_length is None:
            pattern_length = len(pattern)

        self.__is_string = isinstance(pattern, str) or (pattern and isinstance(pattern[0], str))

        # Instantiate the parameters.
        self.__pattern = pattern
        self.__pattern_length = pattern_length

        self.__skip = [self.__pattern_length,] * alphabet_length

        for k in range(self.__pattern_length - 1):
            self.__skip[self.__get_value(pattern[k])] = self.__pattern_length - k - 1

    def __get_value(self, element: Union[str, int]) -> int:
        """
        Return the ID of the element, whether element is a string.
        If element is an integer, return itself or (256 + element) whether it is negative.
        """
        if self.__is_string: return ord(element)
        else: return (256 + element) if element < 0 else element

    def search(self, sequence: Sequence, length: Optional[int] = None) -> Generator[int, None, None]:
        """
        Return all the matching position of pattern.
        """
        if length is None:
            length = len(sequence)

        if self.__pattern_length > length:
            return

        k = self.__pattern_length - 1

        while k < length:
            j = self.__pattern_length - 1
            i = k

            while j >= 0 and sequence[i] == self.__pattern[j]:
                j -= 1
                i -= 1

            if j == -1:
                yield i + 1

            k += self.__skip[self.__get_value(sequence[k])]



# -*- coding: utf-8 -*-

# Read more about process_vm_(read/write)v here:
# https://man7.org/linux/man-pages/man2/process_vm_readv.2.html

# Read more about proc and memory mapping here:
# https://man7.org/linux/man-pages/man5/proc.5.html

from ctypes import addressof, sizeof
from typing import Dict, Generator, Optional, Sequence, Tuple, Type, TypeVar, Union

from ..enums import ScanTypesEnum
from ..util import convert_from_byte_array, get_c_type_of, scan_memory, scan_memory_for_exact_value
from .ptrace import libc
from .types import MEMORY_BASIC_INFORMATION, iovec

import ctypes


T = TypeVar("T")


def get_memory_regions(pid: int) -> Generator[dict, None, None]:
    """
    Generates dictionaries with the address and size of a region used by the process.
    """
    mapping_filename = "/proc/{}/maps".format(pid)

    with open(mapping_filename, "r") as mapping_file:
        for line in mapping_file:

            # Each line keeps information about a memory region of the process.
            region_information = line.split()

            addressing_range, privileges, offset, device, inode = region_information[0: 5]
            path = region_information[5] if len(region_information) >= 6 else str()

            # Convert hexadecimal values to decimal.
            start_address, end_address = [int(addr, 16) for addr in addressing_range.split("-")]
            major_id, minor_id = [int(_id, 16) for _id in device.split(":")]

            offset = int(offset, 16)
            inode = int(inode, 16)

            # Calculate the region size.
            size = end_address - start_address

            region = MEMORY_BASIC_INFORMATION(start_address, size, privileges.encode(), offset, major_id, minor_id, inode, path.encode())
            yield {"address": start_address, "size": region.RegionSize, "struct": region}


def read_process_memory(
    pid: int,
    address: int,
    pytype: Type[T],
    bufflength: int
) -> T:
    """
    Return a value from a memory address.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    data = get_c_type_of(pytype, bufflength)

    libc.process_vm_readv(
        pid, (iovec * 1)(iovec(addressof(data), sizeof(data))),
        1, (iovec * 1)(iovec(address, sizeof(data))), 1, 0
    )
    return data.value.decode() if pytype is str else data.value


def search_addresses_by_value(
    pid: int,
    pytype: Type[T],
    bufflength: int,
    value: Union[bool, int, float, str, bytes, tuple],
    scan_type: ScanTypesEnum = ScanTypesEnum.EXACT_VALUE,
    progress_information: bool = False,
    writeable_only: bool = False,
) -> Generator[Union[int, Tuple[int, dict]], None, None]:
    """
    Search the whole memory space, accessible to the process,
    for the provided value, returning the found addresses.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    # Convert the target value, or all values of a tuple, as bytes.
    target_values = value if isinstance(value, tuple) else (value,)

    conversion_buffer = list()

    for v in target_values:
        target_value = get_c_type_of(pytype, bufflength)
        target_value.value = v.encode() if isinstance(v, str) else v

        target_value_bytes = ctypes.cast(ctypes.byref(target_value), ctypes.POINTER(ctypes.c_byte * bufflength))
        conversion_buffer.append(bytes(target_value_bytes.contents))

    target_value_bytes = tuple(conversion_buffer) if isinstance(value, tuple) else conversion_buffer[0]

    checked_memory_size = 0
    memory_total = 0
    memory_regions = list()

    # Get the memory regions, computing the total amount of memory to be scanned.
    for region in get_memory_regions(pid):

        # Only readable memory pages.
        if b"r" not in region["struct"].Privileges: continue

        # If writeable_only is True, checks if the memory page is writeable.
        if writeable_only and b"w" not in region["struct"].Privileges: continue

        memory_total += region["size"]
        memory_regions.append(region)

    # Sort the list to return ordered addresses.
    memory_regions.sort(key=lambda region: region["address"])

    # Check each memory region used by the process.
    for region in memory_regions:
        address, size = region["address"], region["size"]
        region_data = (ctypes.c_byte * size)()

        # Get data from the region.
        libc.process_vm_readv(
            pid, (iovec * 1)(iovec(addressof(region_data), sizeof(region_data))),
            1, (iovec * 1)(iovec(address, sizeof(region_data))), 1, 0
        )

        # Choose the searching method.
        searching_method = scan_memory

        if scan_type in [ScanTypesEnum.EXACT_VALUE, ScanTypesEnum.NOT_EXACT_VALUE]:
            searching_method = scan_memory_for_exact_value

        # Search the value and return the found addresses.
        for offset in searching_method(region_data, size, target_value_bytes, bufflength, scan_type, pytype is str):
            found_address = address + offset

            extra_information = {
                "memory_total": memory_total,
                "progress": (checked_memory_size + offset) / memory_total,
            }
            yield (found_address, extra_information) if progress_information else found_address

        # Compute the region size to the checked memory size.
        checked_memory_size += size


def search_values_by_addresses(
    pid: int,
    pytype: Type[T],
    bufflength: int,
    addresses: Sequence[int],
    *,
    memory_regions: Optional[Sequence[Dict]] = None,
    raise_error: bool = False,
) -> Generator[Tuple[int, Optional[T]], None, None]:
    """
    Search the whole memory space, accessible to the process,
    for the provided list of addresses, returning their values.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    memory_regions = list(memory_regions) if memory_regions else list()
    addresses = sorted(addresses)

    # If no memory page has been given, get all readable memory pages.
    if not memory_regions:
        for region in get_memory_regions(pid):
            if b"r" not in region["struct"].Privileges: continue
            memory_regions.append(region)

    memory_regions.sort(key=lambda region: region["address"])
    address_index = 0

    # Walk by each memory region.
    for region in memory_regions:
        if address_index >= len(addresses): break

        target_address = addresses[address_index]

        # Check if the memory region contains the target address.
        base_address, size = region["address"], region["size"]
        if not (base_address <= target_address < base_address + size): continue

        region_data = (ctypes.c_byte * size)()

        # Get data from the region.
        libc.process_vm_readv(
            pid, (iovec * 1)(iovec(addressof(region_data), sizeof(region_data))),
            1, (iovec * 1)(iovec(base_address, sizeof(region_data))), 1, 0
        )

        # Get the value of each address.
        while base_address <= target_address < base_address + size:
            offset = target_address - base_address
            address_index += 1

            try:
                data = region_data[offset: offset + bufflength]
                data = (ctypes.c_byte * bufflength)(*data)
                yield target_address, convert_from_byte_array(data, pytype, bufflength)

            except Exception as error:
                if raise_error: raise error
                yield target_address, None

            if address_index >= len(addresses): break
            target_address = addresses[address_index]


def write_process_memory(
    pid: int,
    address: int,
    pytype: Type[T],
    bufflength: int,
    value: Union[bool, int, float, str, bytes]
) -> T:
    """
    Write a value to a memory address.
    """
    if pytype not in [bool, int, float, str, bytes]:
        raise ValueError("The type must be bool, int, float, str or bytes.")

    data = get_c_type_of(pytype, bufflength)
    data.value = value.encode() if isinstance(value, str) else value

    libc.process_vm_writev(
        pid, (iovec * 1)(iovec(addressof(data), sizeof(data))),
        1, (iovec * 1)(iovec(address, sizeof(data))), 1, 0
    )
    return value



# -*- coding: utf-8 -*-
from enum import Enum


class MemoryTypesEnum(Enum):
    """
    Enum with all types of a memory page.
    """
    # Indicates that the memory pages within the region are mapped into the view of an image section.
    MEM_IMAGE = 0x1000000

    # Indicates that the memory pages within the region are mapped into the view of a section.
    MEM_MAPPED = 0x40000

    # Indicates that the memory pages within the region are private (that is, not shared by other processes).
    MEM_PRIVATE = 0x20000



# -*- coding: utf-8 -*-

from typing import Type, TypeVar
import ctypes


T = TypeVar("T")


def convert_from_byte_array(byte_array: ctypes.Array, pytype: Type[T], length: int) -> T:
    """
    Convert a byte array to a Python type.
    """
    if pytype is bytes: return bytes(byte_array)
    if pytype is str: return bytes(byte_array).decode()

    c_value = get_c_type_of(pytype, length)

    return c_value.__class__.from_buffer(byte_array).value


def get_c_type_of(pytype: Type, length) -> ctypes._SimpleCData:
    """
    Return a C type of a primitive type of the Python language.
    """
    if pytype is str or pytype is bytes: return ctypes.create_string_buffer(length)

    elif pytype is int:

        if length == 1: return ctypes.c_int8()      # 1 Byte
        if length == 2: return ctypes.c_int16()     # 2 Bytes
        if length <= 4: return ctypes.c_int32()     # 4 Bytes
        return ctypes.c_int64()                     # 8 Bytes

    # Float values lose their precision when converted to c_float. For that reason,
    # any float value will be converted to double.
    elif pytype is float: return ctypes.c_double()  # 8 Bytes

    elif pytype is bool: return ctypes.c_bool()

    else: raise ValueError("The type must be bool, int, float, str or bytes.")
